import numpy as np
import pylab as plt
from GeneralTmm import Tmm, Material

# Refractive index data of silver from
# P. B. Johnson and R. W. Christy. Optical Constants of the Noble Metals, Phys. Rev. B 6, 4370-4379 (1972)

wlsAg = np.array([4.0e-07,   4.2e-07,   4.4e-07,   4.6e-07, 4.8e-07,   5.0e-07,
    5.2e-07,   5.4e-07, 5.6e-07,   5.8e-07,   6.0e-07,   6.2e-07, 6.4e-07,
    6.6e-07,   6.8e-07,   7.0e-07, 7.2e-07,   7.4e-07,   7.6e-07,   7.8e-07,
    8.0e-07,   8.2e-07,   8.4e-07,   8.6e-07, 8.8e-07,   9.0e-07,   9.2e-07,
    9.4e-07, 9.6e-07,   9.8e-07])
    
nsAg = np.array([0.050 + 2.104j, 0.046 + 2.348j, 0.040 + 2.553j, 0.044 + 2.751j,
    0.050 + 2.948j, 0.050 + 3.131j, 0.050 + 3.316j, 0.057 + 3.505j, 0.057 + 3.679j,
    0.051 + 3.841j, 0.055 + 4.010j, 0.059 + 4.177j, 0.055 + 4.332j, 0.050 + 4.487j,
    0.045 + 4.645j, 0.041 + 4.803j, 0.037 + 4.960j, 0.033 + 5.116j, 0.031 + 5.272j,
    0.034 + 5.421j, 0.037 + 5.570j, 0.040 + 5.719j, 0.040 + 5.883j, 0.040 + 6.048j,
    0.040 + 6.213j, 0.040 + 6.371j, 0.040 + 6.519j, 0.040 + 6.667j, 0.040 + 6.815j,
    0.040 + 6.962j])

if __name__ == "__main__":
    # Prepare materials
    prismN = Material.Static(1.5)
    metalN = Material(wlsAg, nsAg)
    substrateN = Material.Static(1.0)
    
    # Parameters
    wl = 950e-9 # wavelength
    metalD = 50e-9 # metal layer thickness
    betas = np.linspace(0.8, 1.2, 200) # effective mode index
    angles = np.arcsin(betas / prismN(wl).real) # angle inside prism
    xs = np.linspace(-0.5e-6, 2e-6, 300) # for plotting fields
    ys = np.linspace(-1e-6, 1e-6, 301) # for plotting fields
    
    # Enhancement measurement polarisation ((1, 0) = p and (0, 1) = s), layerNr
    # and distance from the interface 
    enhPol, enhLayer, enhDist = np.array([1.0, 0.0]), 2, 0.0 
    enhPos = (enhPol, enhLayer, enhDist)
    
    # Init TMM
    tmm = Tmm(wl = wl)
    
    # Add layers
    tmm.AddIsotropicLayer(float("inf"), prismN)
    tmm.AddIsotropicLayer(metalD, metalN)
    tmm.AddIsotropicLayer(float("inf"), substrateN)
    
    # Reflection, transmission, enhancement
    #--------------------------------------------------------------------------- 
    
    # Do calculations
    sr = tmm.Sweep("beta", betas, enhPos = enhPos)
    
    # Plot
    plt.figure(figsize = (8, 5))
    plt.title("Reflection, transmission and enhancement of surface plasmons on Ag")
    plt.plot(np.degrees(angles), sr["R11"], label = r"reflection")
    plt.plot(np.degrees(angles), sr["T31"], label = r"transmission")
    plt.xlabel(r"$\theta$ ($\degree$)")
    plt.ylabel("reflection/transmission")
    plt.legend()
    plt.twinx()
    plt.plot(np.degrees(angles), sr["enh"], "--", color = "red", label = r"enhancement")
    plt.ylabel("enhencement")
    plt.legend()
    
    # Optimize enhancment to find maximal value (important in case of narrow resonances)
    #--------------------------------------------------------------------------- 
    
    # Set simplex optimizer parameters
    tmm.SetParams(enhOptMaxIters = 100)
    tmm.SetParams(enhOptRel = 1e-5)
    tmm.SetParams(enhInitialStep = 1e-3)
    
    # Initial guess from previous calculations
    maxEnhBetaApprox = betas[np.argmax(sr["enh"])]
    
    # Optimize
    maxEnhOpt = tmm.OptimizeEnhancement(["beta"], np.array([maxEnhBetaApprox]), enhPos)
    maxEnhOptBeta = tmm.beta
    
    # Print result
    print("Initial enhancement %.2f" % (np.max(sr["enh"])))
    print(r"Optimized enhancement %.2f at β=%.4f" % (maxEnhOpt, maxEnhOptBeta))
   
    # Calculate and plot 1D fields at the maximum enhancement angle of incidence
    #--------------------------------------------------------------------------- 
    
    tmm.beta = maxEnhOptBeta
    E1D, H1D = tmm.CalcFields1D(xs, enhPol)
    
    plt.figure()
    plt.title("1D fields of surface plasmons on Ag film")
    plt.plot(1e6 * xs, abs(E1D[:, 0]), label = r"|$E_x$|")
    plt.plot(1e6 * xs, abs(E1D[:, 1]), label = r"|$E_y$|")
    plt.plot(1e6 * xs, abs(E1D[:, 2]), label = r"|$E_z$|")
    plt.plot(1e6 * xs, np.linalg.norm(E1D, axis = 1), "--", label = r"|$E$|")
    plt.axvline(0.0, ls = "--", color = "black", lw = 1.0)
    plt.axvline(1e6 * metalD, ls = "--", color = "black", lw = 1.0)
    plt.xlabel("x ($\mu m$)")
    plt.ylabel("Electrical field (V/m)")
    plt.legend()
    
    # Calculate and plot 2D fields at the maximum enhancement angle of incidence
    #--------------------------------------------------------------------------- 
    
    tmm.beta = maxEnhOptBeta
    E2D, H2D = tmm.CalcFields2D(xs, ys, enhPol)
    
    toPlot = [(r"$E_x$ (V/m)", E2D[:, :, 0].real),
              (r"$E_y$ (V/m)", E2D[:, :, 1].real),
              (r"$H_z$ (A/m)", H2D[:, :, 2].real),
              (r"$|E|$ (V/m)", np.linalg.norm(E2D, axis = 2))]
    
    plt.figure()
    plt.suptitle("2D fields of surface plasmons on Ag film")
    for i, (label, data) in enumerate(toPlot):
        plt.subplot(221 + i)
        plt.pcolormesh(1e6 * xs, 1e6 * ys, data.real.T)
        plt.axvline(0.0, ls = "--", color = "white", lw = 0.5)
        plt.axvline(1e6 * metalD, ls = "--", color = "white", lw = 0.5)
        plt.xlabel("x ($\mu m$)")
        plt.ylabel("y ($\mu m$)")
        cbar = plt.colorbar()
        cbar.set_label(label)
    plt.tight_layout()
    plt.subplots_adjust(top = 0.92)
    
    plt.show()


import numpy as np
from libcpp.pair cimport pair
from libcpp.vector cimport vector
from CppGeneralTmm cimport *

#===============================================================================
# Common
#===============================================================================

cdef enum ParamDatatype:
    PARAM_DT_DOUBLE,
    PARAM_DT_INT,
    PARAM_DT_COMPLEX

cdef pair[ParamCpp, ParamDatatype] ToCppParam(str param) except +: 
    """ToCppParam(param)
    
    Converts the name of the param (e.g wl, beta, ...) to C++ param class.
    
    Parameters
    ----------
    param : str
        The name of the parameter.
        
    Returns
    -------
    tuple of ParamCpp and ParamDatatype
        C++ class describing the parameter and its datatype
        
    """
    cdef int layerNr = -1
    if param.count("_") == 1:
        layerNr = int(param.split("_")[-1])
    elif param.count("_") > 1:
        raise ValueError("Unknown param: %s" % (param))

    if param == "wl":
        return pair[ParamCpp, ParamDatatype](ParamCpp(WL), PARAM_DT_DOUBLE)
    elif param == "beta":
        return pair[ParamCpp, ParamDatatype](ParamCpp(BETA), PARAM_DT_DOUBLE)
    elif param == "enhOptRel":
        return pair[ParamCpp, ParamDatatype](ParamCpp(ENH_OPT_REL), PARAM_DT_DOUBLE)
    elif param == "enhOptMaxIters":
        return pair[ParamCpp, ParamDatatype](ParamCpp(ENH_OPT_MAX_ITERS), PARAM_DT_INT)
    elif param == "enhInitialStep":
        return pair[ParamCpp, ParamDatatype](ParamCpp(ENH_INITIAL_STEP), PARAM_DT_DOUBLE)
    elif param.startswith("d_"):
        return pair[ParamCpp, ParamDatatype](ParamCpp(LAYER_D, layerNr), PARAM_DT_DOUBLE)
    elif param.startswith("n_"):
        return pair[ParamCpp, ParamDatatype](ParamCpp(LAYER_N, layerNr), PARAM_DT_COMPLEX)
    elif param.startswith("nx_"):
        return pair[ParamCpp, ParamDatatype](ParamCpp(LAYER_NX, layerNr), PARAM_DT_COMPLEX)
    elif param.startswith("ny_"):
        return pair[ParamCpp, ParamDatatype](ParamCpp(LAYER_NY, layerNr), PARAM_DT_COMPLEX)
    elif param.startswith("nz_"):
        return pair[ParamCpp, ParamDatatype](ParamCpp(LAYER_NZ, layerNr), PARAM_DT_COMPLEX)
    elif param.startswith("psi_"):
        return pair[ParamCpp, ParamDatatype](ParamCpp(LAYER_PSI, layerNr), PARAM_DT_DOUBLE)
    elif param.startswith("xi_"):
        return pair[ParamCpp, ParamDatatype](ParamCpp(LAYER_XI, layerNr), PARAM_DT_DOUBLE)
    else:
        raise ValueError("Unknown param: %s" % (param))

cdef WaveDirectionCpp WaveDirectionFromStr(str waveStr) except +:
    """WaveDirectionFromStr(waveStr)
    
    Helper function to convert wave direction string to WaveDirectionCpp enum.
    
    Parameters
    ----------
    waveStr : str
        Wave direction string (forward, backward or both)
        
    Returns
    -------
    WaveDirectionCpp
        Enum indicating the direction.
    
    """
    if waveStr == "forward":
        return WD_FORWARD
    elif waveStr == "backward":
        return WD_BACKWARD
    elif waveStr == "both":
        return WD_BOTH
    else:
        raise NotImplementedError()


cdef class _SweepRes:
    """_SweepRes()
    
    Helper class to hold the results of the :any:`Tmm.Sweep` method. The
    __getitem__ method is overloaded to return the results stored in the
    internal dictionary. For full details of the calculated quantities see the
    publication Hodgkinson, I. J., Kassam, S., & Wu, Q. H. (1997) Journal of 
    Computational Physics, 133(1) 75-83
    
    Attributes
    ----------
    ["R11"] : ndarray of floats
        The intensity reflection coefficient of p-polarized wave (1 == p).
    ["R22"] : ndarray of floats
        The intensity reflection coefficient of s-polarized wave (2 == s).
    ["R12"] : ndarray of floats
        The anisotropic reflection intensity coefficient. Incident light is
        s-polarized (2) and the measured field is p-polarized(1).
    ["R21"] : ndarray of floats
        The anisotropic reflection intensity coefficient. Incident light is
        p-polarized (1) and the measured field is s-polarized(2).
    ["T31"] : ndarray of floats
        The intensity transmission coefficient of p-polarized wave (1 & 3 == p).
    ["T42"] : ndarray of floats
        The intensity transmission coefficient of s-polarized wave (2 & 4 == s).
    ["T32"] : ndarray of floats
        The anisotropic transmission intensity coefficient. Incident light is
        s-polarized (2) and the measured field is p-polarized (3).
    ["T41"] : ndarray of floats
        The anisotropic transmission intensity coefficient. Incident light is
        p-polarized (1) and the measured field is s-polarized (4).
    ["rXY"] : ndarray of complex
        The amplitude reflection coefficients similarly to intensity coefficients.
        The X denotes measured field polarization (1..4) and Y incident
        polarization (1..4)
    ["tXY"] : ndarray of complex
        The amplitude transmission coefficients similarly to intensity coefficients.
        The X denotes measured field polarization (1..4) and Y incident
        polarization (1..4)
    ["enh"] : ndarray of floats
        The the enhancement of electrical field norm in specified position in
        comparison to the incident field norm in vacuum.
    ["enhEx"] : ndarray of floats
        The the enhancement of electrical field x-component amplitude in
        specified position in comparison to the incident field norm in vacuum.
    ["enhEy"] : ndarray of floats
        The the enhancement of electrical field y-component amplitude in
        specified position in comparison to the incident field norm in vacuum.
    ["enhEz"] : ndarray of floats
        The the enhancement of electrical field z-component amplitude in
        specified position in comparison to the incident field norm in vacuum.
        
    """
    cdef dict _res;

    def __init__(self):
        self._res = {}
    
    cdef _Init(self, SweepResCpp resCpp):
        cdef map[string, ArrayXcd] mapComplex = resCpp.mapComplex
        cdef map[string, ArrayXd] mapDouble = resCpp.mapDouble
        
        for kvPair in mapComplex:
            self._res[kvPair.first.decode()] = ndarray_copy(kvPair.second).squeeze()
        for kvPair2 in mapDouble:
            self._res[kvPair2.first.decode()] = ndarray_copy(kvPair2.second).squeeze() 

    def __getitem__(self, index):
        return self._res[index]
    
    def __repr__(self):
        return "%s(%r)" % (self.__class__, self._res)


#===============================================================================
# Material
#===============================================================================

cdef class Material:
    """Material(wls, ns)
    
    This class describes the optical parameters of medium. Default
    contructor takes arrays of wavelengths and complex refractive indices and
    does linear interpolation. For shortcut __call__(wl) is defined as GetN(wl).
    
    Parameters
    ----------
    wls : ndarray of floats
        Array of wavelengths (m)
    ns : ndarray of complex floats
        Corresponding complex refractive indices to the wls array.
        
    """
    cdef MaterialCpp *_thisptr
    cdef readonly object chi2

    def __cinit__(self, np.ndarray[double, ndim = 1] wls, np.ndarray[double complex, ndim = 1] ns):
        # Copy is made in c++
        self._thisptr = new MaterialCpp(Map[ArrayXd](wls), Map[ArrayXcd](ns))
                
    def __dealloc__(self):
        del self._thisptr
        
    def __call__(self, wl): # For compatibility
        return self.GetN(wl)
        
    def GetN(self, wl):
        """GetN(wl)
        
        Returns refractive index of material at specified wavelength.
        
        Parameters
        ----------
        wl : float
            Wavelength in meters.
        
        Returns
        -------
        complex
            Complex refractive index at wavelength wl.
        
        Examples
        --------
        >>> wls = np.array([400e-9, 600e-9, 800e-9], dtype = float)
        >>> ns = np.array([1.5 + 0.3j, 1.7 + 0.2j, 1.8 + 0.1j], dtype = complex)
        >>> mat = Material(wls, ns)
        >>> mat(600e-9)
        1.7 + 0.2j
        
        """
        return self._thisptr.n(wl)


#===============================================================================
# Tmm
#===============================================================================

cdef class Tmm:
    """Tmm(**kwargs)
    
    This is main class for 4x4 matrix anisotropic transfer-matrix method (TMM).
    Allows to calculate both isotropic and anisotropic mediums. In case of
    anisotropic mediums the layer is specified by the thickness, by three
    refractive indices (along each axis) and by to angles denoting the alignment
    of crystallographic axes with structure axes. This class is wrapper for the
    code written in C++. This implementation is based on the publication
    Hodgkinson, I. J., Kassam, S., & Wu, Q. H. (1997). Journal of Computational
    Physics, 133(1) 75-83.
    
    Parameters
    ----------
    **kwargs : dict
        All parameters are passed to :any:`Tmm.SetParams`.
     
    Attributes
    ----------
    wl : float
        Wavelength in meters.
    beta : float
        Effective mode index. Determines the angle of incidence.
    enhOptRel : float
        The relative error termination condition for optimization.
    enhOptMaxIters : int
        Maximum number of optimization iterations.
    enhInitialStep : float
        Initial step for optimization.
     
    """
    cdef TmmCpp *_thisptr
    cdef readonly list materialsCache
    
    def __cinit__(self, **kwargs):
        self._thisptr = new TmmCpp()
        self.materialsCache = []
        self.SetParams(**kwargs)
        
    def __dealloc__(self):
        if self._thisptr is not NULL:
            del self._thisptr

    def SetParams(self, **kwargs):
        """SetParams(**kwargs)
        
        Convenience function to set multiple parameters at once. For the list
        of parameters see function :any:`ToCppParam`
        
        Parameters
        ----------
        **kwargs : dict
            Param and value pais to set.
        
        """
        cdef pair[ParamCpp, ParamDatatype] paramDef
        for name, value in kwargs.items():
            paramDef = ToCppParam(name)
            if paramDef.second == PARAM_DT_DOUBLE:
                self._thisptr.SetParam(paramDef.first, <double>value)
            elif paramDef.second == PARAM_DT_INT:
                self._thisptr.SetParam(paramDef.first, <int>value)
            elif paramDef.second == PARAM_DT_COMPLEX:
                self._thisptr.SetParam(paramDef.first, <double complex>value)                
            else:
                raise ValueError("Unknown param datatype.")
        
    def AddIsotropicLayer(self, double d, Material mat):
        """AddIsotropicLayer(d, mat)
        
        Adds isotropic layer to transfer-matrix method.
        
        Parameters
        ----------
        d : float
            The thickness of the layer. First and last layers should have
            thikness equal to float("inf").
        mat : :any:`Material`
            Instance of the :any:`Material` class decribing the refractive
            index dispersion over wavelength. 
        
        """
        self._thisptr.AddIsotropicLayer(d, mat._thisptr)
        self.materialsCache.append(mat) # Avoids dealloc

    def AddLayer(self, double d, Material matx, Material maty, Material matz, double psi, double xi):
        """AddLayer(d, mat)
        
        Adds anisotropic layer to transfer-matrix method.
        
        Parameters
        ----------
        d : float
            The thickness of the layer. First and last layers should have
            thikness equal to float("inf").
        matx : :any:`Material`
            Instance of the :any:`Material` class decribing the refractive
            index dispersion over wavelength in x-direction.
        maty : :any:`Material`
            Instance of the :any:`Material` class decribing the refractive
            index dispersion over wavelength in y-direction.
        matz : :any:`Material`
            Instance of the :any:`Material` class decribing the refractive
            index dispersion over wavelength in z-direction.
        psi : float
            The rotational angle around the z-axis for alignment of the material.
            Measured in radians.
        xi : float
            The rotational angle around the x-axis for alignment of the material.
            Measured in radians.
            
        """
        self._thisptr.AddLayer(d, matx._thisptr, maty._thisptr, matz._thisptr, psi, xi)
        self.materialsCache.append(matx) # Avoids dealloc
        self.materialsCache.append(maty) # Avoids dealloc
        self.materialsCache.append(matz) # Avoids dealloc

    def ClearLayers(self):
        """ClearLayers()
        
        Clears all the layers from the structure.
        
        """
        
        self._materialsCache.clear()
        self._thisptr.ClearLayers()

    def GetIntensityMatrix(self):
        """GetIntensityMatrix()
        
        Solves the system and returns the intensity matrix defined by Eq. 20
        in I. J. Hodgkinson, et al (1997). Contains all intensity reflection and
        transmission coefficients. See also :any:`_SweepRes`.
        
        Returns
        -------
        ndarray of floats
            Has shape (4, 4) and contains the intensity matrix coefficients.
        
        """
        return ndarray_copy(self._thisptr.GetIntensityMatrix())
    
    def GetAmplitudeMatrix(self):
        """GetAmplitudeMatrix()
        
        Solves the system and returns the amplitude matrix defined by Eq. 19
        in I. J. Hodgkinson, et al (1997). See also :any:`_SweepRes`.
        
        Returns
        -------
        ndarray of complex floats
            Has shape (4, 4) and contains the amplitude matrix coefficients.
        
        """
        return ndarray_copy(self._thisptr.GetAmplitudeMatrix())
    
    def Sweep(self, paramName, np.ndarray[double, ndim = 1] values, enhPos = None, int alphaLayer = -1):
        """Sweep(paramName, values, enhPos = None, alphaLayer = -1)
        
        Convenience function for solving system for multiple values of a single
        parameter (given by :any:`paramName`).
        
        Parameters
        ----------
        paramName : str
            The name of parameter. See :any:`Tmm.SetParams` for the list of all
            possible parameter names.
        values : ndarray of floats
            The list of values for parameter to solve for.
        enhPos : tuple
            Tuple structured like ((polCoef1, polCoef2), layerNr, distance).
            If this parameter is not None, then the field enhancment will be
            calculated in layer numbered by `layerNr` at distance `distance`
            from the interface. The variables (polCoef1, polCoef2) define the
            polarization for the field enhancement calculation (in comparison to
            the excitation in vacuum) and (1.0, 0.0) corresonds to p-polarization
            and (0.0, 1.0) corresond to s-polarization.
        alphaLayer: int
            Computes the perpendicular component of the wavevectors in layer
            nr `alphaLayer`.
            
        Returns
        -------
        :any:`_SweepRes`
            Helper class containing all the results of the sweep.
        
        """
        cdef pair[ParamCpp, ParamDatatype] paramDef = ToCppParam(paramName)
        cdef PositionSettingsCpp enhPosCpp
        cdef SweepResCpp resCpp
        
        if enhPos is not None:
            enhPosCpp = PositionSettingsCpp(enhPos[0][0], enhPos[0][1], enhPos[1], enhPos[2])
        
        resCpp = self._thisptr.Sweep(paramDef.first, Map[ArrayXd](values), enhPosCpp, <int>alphaLayer)
        res = _SweepRes()
        res._Init(resCpp)
        return res
    
    def CalcFields1D(self, np.ndarray[double, ndim = 1] xs, np.ndarray[double, ndim = 1] polarization, str waveDirectionStr = "both"):
        """CalcFields1D(xs, polarization, waveDirection = "both")
        
        Calculates the complex electric and magnetic fields along the x-axis.
        
        Paramaters
        ----------
        xs : ndarray of floats
            Position array where to calculate the fields.
        polarization : ndarray of floats (2,)
            Contains the polarization coefficients of the excitation.
        waveDirection : str {"both", "forward", "backward"}
            Allows to select the output fields.
            
        Returns
        -------
        tuple (E, H) of ndarrays of complex floats
           Variable E contains the electrical fields and has shape (N, 3), 
           where N is the length of `xs` array and 3 correspond to x-, y- and
           z-direction. Variable H contains magnetic fields in similar manner.  
        
        """
        cdef WaveDirectionCpp waveDirection = WaveDirectionFromStr(waveDirectionStr) 
        cdef EMFieldsListCpp resCpp        
        resCpp = self._thisptr.CalcFields1D(Map[ArrayXd](xs), Map[Array2d](polarization), waveDirection)

        E = ndarray_copy(resCpp.E)
        H = ndarray_copy(resCpp.H)
        return E, H
    
    def CalcFields2D(self, np.ndarray[double, ndim = 1] xs, np.ndarray[double, ndim = 1] ys, np.ndarray[double, ndim = 1] pol, str waveDirectionStr = "both"):
        """CalcFields2D(xs, ys, polarization, waveDirection = "both")
        
        Calculates the complex electric and magnetic fields along in xy-plane in
        the rectangular grid following from `xs` and `ys`.
        
        Paramaters
        ----------
        xs : ndarray of floats
            Position array where to calculate the fields in x-direction.
        ys : ndarray of floats
            Position array where to calculate the fields in y-direction.
        polarization : ndarray of floats (2,)
            Contains the polarization coefficients of the excitation.
        waveDirection : str {"both", "forward", "backward"}
            Allows to select the output fields.
            
        Returns
        -------
        tuple (E, H) of ndarrays of complex floats
           Variable E contains the electrical fields and has shape (N, M, 3), 
           where N is the length of `xs` array, M is the lenght of `ys`
           array and 3 correspond to x-, y- and z-direction. Variable H
           contains magnetic fields in similar manner.  
        
        """
        ky = self.beta * 2.0 * np.pi / self.wl
        phaseY = np.exp(1.0j * ky * ys)
        E1D, H1D = self.CalcFields1D(xs, pol, waveDirectionStr)
        
        E = np.zeros((len(xs), len(ys), 3), dtype = complex)
        H = np.zeros((len(xs), len(ys), 3), dtype = complex)
        for i in range(3):
            E[:, :, i] = np.outer(E1D[:, i], phaseY)
            H[:, :, i] = np.outer(H1D[:, i], phaseY)
        
        return E, H
        
    def CalcFieldsAtInterface(self, enhPos, str waveDirectionStr = "both"):
        """CalcFieldsAtInterface(enhPos, waveDirectionStr = "both")
        
        Calculates electic and magnetic fields in one point defined by
        `enhPos`.
        
        Parameters
        ----------
        enhPos : tuple
            Tuple structured like ((polCoef1, polCoef2), layerNr, distance).
            If this parameter is not None, then the field enhancment will be
            calculated in layer numbered by `layerNr` at distance `distance`
            from the interface. The variables (polCoef1, polCoef2) define the
            polarization for the field enhancement calculation (in comparison to
            the excitation in vacuum) and (1.0, 0.0) corresonds to p-polarization
            and (0.0, 1.0) corresond to s-polarization.
        waveDirection : str {"both", "forward", "backward"}
            Allows to select the output fields.
            
        Returns
        -------
        tuple (E, H) of ndarrays of complex
            Variable E is ndarray of complex floats (length 3) containing
            field x-, y- and z-component. Variable H contains magnetic fields
            in similar manner.
            
        """
        cdef WaveDirectionCpp waveDirection = WaveDirectionFromStr(waveDirectionStr)
        cdef PositionSettingsCpp enhPosCpp
        cdef EMFieldsCpp resCpp

        enhPosCpp = PositionSettingsCpp(enhPos[0][0], enhPos[0][1], enhPos[1], enhPos[2])
        resCpp = self._thisptr.CalcFieldsAtInterface(enhPosCpp, waveDirection)
        
        E = ndarray_copy(resCpp.E).squeeze()
        H = ndarray_copy(resCpp.H).squeeze()
        return E, H
        
    def OptimizeEnhancement(self, list optParams, np.ndarray[double, ndim = 1] optInitials, enhpos):
        """OptimizeEnhancement(optParams, optInitials, enhpos)
        
        Function for optimizing structure params for maximal field enhancment in
        one point (defined by enhpos). See also parameters for optimization:
        `enhOptRel`, `enhOptMaxIters` and `enhInitialStep`.
        
        Parameters
        ----------
        optParams: list of str
            List of parameter names to optimize.
        optInitials: ndarray of float
            Contains initial quesses for parameters in `optParams`.
        enhPos : tuple
            Tuple structured like ((polCoef1, polCoef2), layerNr, distance).
            The field enhancment will be calculated in layer numbered by
            `layerNr` at distance `distance` from the interface. The variables 
            (polCoef1, polCoef2) define the polarization for the field
            enhancement calculation (in comparison to the excitation in vacuum):
            (1.0, 0.0) corresonds to p-polarization and (0.0, 1.0) corresond to
            s-polarization.
        
        Returns
        -------
        float
            The optimized field enhancment. All the parameters will be assigned
            their optimal values and could be thus accessed in usual way.
        
        """
        cdef PositionSettingsCpp enhPosCpp
        cdef vector[ParamCpp] paramVec
        (pol, interface, dist) = enhpos
        enhPosCpp = PositionSettingsCpp(pol[0], pol[1], interface, dist)
        
        for paramName in optParams:
            paramVec.push_back(ToCppParam(paramName).first)
        
        res = self._thisptr.OptimizeEnhancement(paramVec, Map[ArrayXd](optInitials), enhPosCpp)
        return res
    
        
    # Getters
    #--------------------------------------------------------------------------- 
    
    @property
    def wl(self):
        return self._thisptr.GetParamDouble(ParamCpp(WL))
    
    @property
    def beta(self):
        return self._thisptr.GetParamDouble(ParamCpp(BETA))
    
    @property
    def enhOptRel(self):
        return self._thisptr.GetParamDouble(ParamCpp(ENH_OPT_REL))
    
    @property
    def enhOptMaxIters(self):
        return self._thisptr.GetParamInt(ParamCpp(ENH_OPT_MAX_ITERS))
    
    @property
    def enhInitialStep(self):
        return self._thisptr.GetParamDouble(ParamCpp(ENH_INITIAL_STEP))
    
    # Setters
    #--------------------------------------------------------------------------- 
    
    @wl.setter
    def wl(self, double value):  # @DuplicatedSignature
        self._thisptr.SetParam(ParamCpp(WL), <double>value)
        
    @beta.setter
    def beta(self, double value):  # @DuplicatedSignature
        self._thisptr.SetParam(ParamCpp(BETA), <double>value)
    
    @enhOptRel.setter
    def enhOptRel(self, double value):  # @DuplicatedSignature
        self._thisptr.SetParam(ParamCpp(ENH_OPT_REL), <double>value)
    
    @enhOptMaxIters.setter
    def enhOptMaxIters(self, int value):  # @DuplicatedSignature
        self._thisptr.SetParam(ParamCpp(ENH_OPT_MAX_ITERS), <int>value)
    
    @enhInitialStep.setter
    def enhInitialStep(self, double value):  # @DuplicatedSignature
        self._thisptr.SetParam(ParamCpp(ENH_INITIAL_STEP), <double>value)
    
    


"""This module contains classes for 4x4 TMM
Hodgkinson, I. J., Kassam, S., & Wu, Q. H. (1997). 
Journal of Computational Physics, 133(1) 75-83

This Python code is only used for testing the C++ code of the same algorithm.

"""

import numpy as np
import math
import warnings
from scipy import optimize

__all__ = ["TmmPy"]

def RotationSx(phi):
    res = np.array([[1.0, 0.0, 0.0], \
                    [0.0, math.cos(phi), -math.sin(phi)], \
                    [0.0, math.sin(phi), math.cos(phi)]])
    return res

def RotationSz(phi):
    res = np.array([[math.cos(phi), -math.sin(phi), 0.0], \
                    [math.sin(phi), math.cos(phi), 0.0], \
                    [0.0, 0.0, 1.0]])
    return res

def Norm(vector):
    if len(vector.shape) > 1:
        if vector.shape[1] != 3:
            raise Exception("Only vectors with length 3 supported.")
        return np.sqrt(abs(vector[:, 0]) ** 2.0 + abs(vector[:, 1]) ** 2.0 +  abs(vector[:, 2]) ** 2.0, dtype = complex).real
    else:        
        if len(vector) != 3:
            raise Exception("Only vectors with length 3 supported.")
        return np.sqrt(abs(vector[0]) ** 2.0 + abs(vector[1]) ** 2.0 + abs(vector[2]) ** 2.0, dtype = complex).real


class _AnisotropicLayer():
    
    def __init__(self, tmm, d, n1, n2, n3, psi, xi):
        self.tmm = tmm
        self.d = d
        self.n1 = n1
        self.n2 = n2
        self.n3 = n3
        self.psi = psi
        self.xi = xi
        
    def GetConf(self):
        if self.n1 == self.n2 and self.n1 == self.n3:
            return ("iso", self.d, self.n1)
        else:
            return ("aniso", self.d, self.n1, self.n2, self.n3, self.psi, self.xi)
        
    def SetConf(self, **kwargs):
        self.d = kwargs.pop("d", self.d)
        self.n1 = kwargs.pop("n1", self.n1)
        self.n2 = kwargs.pop("n2", self.n2)
        self.n3 = kwargs.pop("n3", self.n3)
        self.psi = kwargs.pop("psi", self.psi)
        self.xi = kwargs.pop("xi", self.xi)
        
    def _GetTangentialFields(self, Ey, Hz, Ez, Hy):
        z0 = 119.9169832 * math.pi
    
        beta = self.tmm.beta
        epsXX = self.epsTensor[0, 0]
        epsXY = self.epsTensor[0, 1]
        epsXZ = self.epsTensor[0, 2]
        
        Ex = -(epsXY * Ey + epsXZ * Ez + beta * z0 * Hz) / epsXX
        Hx = (beta / z0) * Ez
        
        return Ex, Hx
        
    def GetFields(self, x, coefs):
        E = np.array([0.0, 0.0, 0.0], dtype = complex)
        H = np.array([0.0, 0.0, 0.0], dtype = complex)
        for mode in range(4):
            a = self.alpha[mode]
         
            mEy, mHz, mEz, mHy = coefs[mode] * self.F[:, mode]
            mEx, mHx = self._GetTangentialFields(mEy, mHz, mEz, mHy)
            
            phase = np.exp(1.0j * self.tmm.k0 * a * x)
            
            E[0] += mEx * phase
            E[1] += mEy * phase
            E[2] += mEz * phase
            H[0] += mHx * phase
            H[1] += mHy * phase
            H[2] += mHz * phase
        
        return E, H
    
        
    def Solve(self, beta):
        self._CalcEpsilonMatrix()
        self._SolveEigenFunction(beta)
        self._SolvePhaseMatrix()
        self._SolveFieldTransferMatrix()
            
    
    def _CalcEpsilonMatrix(self):
                
        self.epsTensorCrystal = np.array([[self.n1 ** 2.0, 0.0, 0.0], \
                                          [0.0, self.n2 ** 2.0, 0.0], \
                                          [0.0, 0.0, self.n3 ** 2.0]], dtype = complex)
        
        
        
        self.epsTensor = np.dot(RotationSx(self.xi), \
                                np.dot(RotationSz(self.psi), \
                                np.dot(self.epsTensorCrystal, \
                                np.dot(RotationSz(-self.psi), RotationSx(-self.xi))))) 

        

    def _SolveEigenFunction(self, beta):
        z0 = 119.9169832 * math.pi
    
        epsXX = self.epsTensor[0, 0]
        epsYY = self.epsTensor[1, 1]
        epsZZ = self.epsTensor[2, 2]
        epsXY = self.epsTensor[0, 1]
        epsXZ = self.epsTensor[0, 2]
        epsYZ = self.epsTensor[1, 2]
        
        #print
        #print "Solve Eigen Function", beta
        
        #print "epsTensor"
        #print self.epsTensor
        
        mBeta = np.zeros((4, 4), dtype = complex)
        mBeta[0, 0] = -beta * epsXY / epsXX
        mBeta[0, 1] = z0 - (z0 * beta ** 2.0) / epsXX
        mBeta[0, 2] = -beta * epsXZ / epsXX
        mBeta[0, 3] = 0.0
        mBeta[1, 0] = epsYY / z0 - (epsXY ** 2.0) / (z0 * epsXX)
        mBeta[1, 1] = (-beta * epsXY) / epsXX
        mBeta[1, 2] = epsYZ / z0 - (epsXY * epsXZ) / (z0 * epsXX)
        mBeta[1, 3] = 0.0
        mBeta[2, 0] = 0.0
        mBeta[2, 1] = 0.0
        mBeta[2, 2] = 0.0
        mBeta[2, 3] = -z0
        mBeta[3, 0] = (-epsYZ / z0) + (epsXY * epsXZ) / (z0 * epsXX)
        mBeta[3, 1] = beta * epsXZ / epsXX
        mBeta[3, 2] = (beta ** 2.0) / z0 + (epsXZ ** 2.0) / (z0 * epsXX) - epsZZ / z0
        mBeta[3, 3] = 0.0
        values, vectors = np.linalg.eig(mBeta)

        
        poyntingX = np.zeros((4), dtype = float)
        for i in range(4):
            poyntingX[i] = 0.5 * (vectors[0, i] * np.conj(vectors[1, i]) - \
                                  vectors[2, i] * np.conj(vectors[3, i])).real
        
        
        forward, backward = [], []
        for i in range(4):
            movingForward = False
            if abs(poyntingX[i]) > 1e-10:
                movingForward = poyntingX[i] > 0.0
            else:
                movingForward = values[i].imag > 0.0
            
            if movingForward:
                forward.append(i)
            else:
                backward.append(i)
            
        if len(forward) != 2:
            print("ns", self.n1, self.n2, self.n3)
            print("beta", beta)
            print("Values", values)
            print("Poynting", poyntingX)
            print("vectors", vectors)
            raise Exception("Wrong number of forward moving waves: %d" % len(forward))
        
        if abs(values.real[forward[0]] - values.real[forward[1]]) < 1e-10:
            #print "TODDO", values[forward[0]], values[forward[1]]
            pass
        elif values.real[forward[0]] < values.real[forward[1]]:
            forward[0], forward[1] = forward[1], forward[0]
        
        if abs(values.real[backward[0]] - values.real[backward[1]]) < 1e-10:
            #print "TODDO BACK", values[backward[0]], values[backward[1]]
            pass
        elif values.real[backward[0]] > values.real[backward[1]]:
            backward[0], backward[1] = backward[1], backward[0]
        

        option1 = [forward[0], backward[0], forward[1], backward[1]]
        #option2 = [forward[0], backward[1], forward[1], backward[0]]
        order = option1

        self.alpha = values[order]
        self.F = vectors[ :, order]
        self.poynting = poyntingX[order]
        self.invF = np.linalg.inv(self.F)
   

        
    def _SolvePhaseMatrix(self):
        self.phaseMatrix = np.identity(4, dtype = complex)
        
        if self.d == float("inf"):
            return
        
        for i in range(4):
            phi = self.tmm.k0 * self.alpha[i] * self.d
            self.phaseMatrix[i, i] = np.exp(-1.0j * phi)
    
    def _SolveFieldTransferMatrix(self):
        self.M = np.dot(self.F, np.dot(self.phaseMatrix, self.invF))
        
    
class TmmPy():
    
    def __init__(self):
        self.wl = None
        self.beta = None
        self.polarization = None
        self.layers = []
        self.namesr = [["r11", "r12", "t13", "t14"], \
                       ["r21", "r22", "t23", "t24"], \
                       ["t31", "t32", "r33", "r34"], \
                       ["t41", "t42", "r43", "r44"]]
        
        self.namesR = [["R11", "R12", "T13", "T14"], \
                       ["R21", "R22", "T23", "T24"], \
                       ["T31", "T32", "R33", "R34"], \
                       ["T41", "T42", "R43", "R44"]]
        
    def AddIsotropicLayer(self, d, n):
        self.layers.append(_AnisotropicLayer(self, d, n, n, n, 0.0, 0.0))
        
    def AddLayer(self, d, n1, n2, n3, psi, xi):
        self.layers.append(_AnisotropicLayer(self, d, n1, n2, n3, psi, xi))
    
    def GetConf(self):
        layersConf = [layer.GetConf() for layer in self.layers]
        res = {"wl": self.wl, "beta": self.beta, \
               "polarization": self.polarization, "layers": layersConf}    
        return res
    
    def SetConf(self, **kwargs):
        self.wl = kwargs.pop("wl", self.wl)
        self.beta = kwargs.pop("beta", self.beta)
        self.polarization = kwargs.pop("polarization", self.polarization)
        
        # All layer params
        layersConf = kwargs.pop("layers", None)
        if layersConf != None:
            self.layers = []
            for lConf in layersConf:
                if lConf[0] == "iso":
                    _, d, n = lConf
                    self.AddIsotropicLayer(d, n)
                elif lConf[0] == "aniso":
                    _, d, n1, n2, n3, psi, xi = lConf
                    self.AddLayer(d, n1, n2, n3, psi, xi)
                else:
                    raise NotImplemented()
        
        # Layer individual params
        for key, value in list(kwargs.items()):
            if key.find("_") == -1:
                continue
            kwargs.pop(key)
            param, index = key.split("_")
            index = int(index)
            self.layers[index].SetConf(**{param: value})

    
    def SolveFor(self, param, values, polarization = None, enhInterface = None, enhDist = 0.0):
        if polarization != None:
            self.polarization = polarization
            
        res = {}
        for i in range(4):
            for j in range(4):
                res[self.namesr[i][j]] = np.zeros_like(values, dtype = complex)
                res[self.namesR[i][j]] = np.zeros_like(values, dtype = float)
                
        if enhInterface != None:
            res["enh"] = np.zeros_like(values, dtype = float)
                
        for i in range(len(self.layers)):
            res["alphas_%d" % (i)] = np.zeros((len(values), 4), dtype = complex)
                
                
        for i in range(len(values)):
            self.SetConf(**{param: values[i]})
            r, R = self.Solve()
            
            if enhInterface != None: 
                res["enh"][i], _ = self.CalcEnhAtInterface(None, enhInterface, enhDist = enhDist)
    
                
            for j in range(len(self.layers)):
                res["alphas_%d" % (j)][i] = self.layers[j].alpha[:]
    
            #if R[0, 0] > 1.0 + 1e-6:
                #print R

            for j in range(4):
                for k in range(4):
                    res[self.namesr[j][k]][i] = r[j, k]
                    res[self.namesR[j][k]][i] = R[j, k]

        return res

    def Solve(self, wl = None, beta = None):
        if wl == None: wl = self.wl
        if beta == None: beta = self.beta
        self.wl = wl
        self.beta = beta
        self.k0 = 2.0 * math.pi / wl
        
        for layer in self.layers:
            layer.Solve(beta)
            
        # System matrix
        self.A = self.layers[0].invF
        for i in range(1, len(self.layers) - 1):
            self.A = np.dot(self.A, self.layers[i].M)
            
        
            
        self.A = np.dot(self.A, self.layers[-1].F)

        
        # r-matrix
        A = self.A
        r1 = np.array([[0.0, 0.0, -A[0, 0], -A[0, 2]], \
                       [1.0, 0.0, -A[1, 0], -A[1, 2]], \
                       [0.0, 0.0, -A[2, 0], -A[2, 2]], \
                       [0.0, 1.0, -A[3, 0], -A[3, 2]]], dtype = complex)
        r2 = np.array([[-1.0, 0.0, A[0, 1], A[0, 3]], \
                       [0.0, 0.0, A[1, 1], A[1, 3]], \
                       [0.0, -1.0, A[2, 1], A[2, 3]], \
                       [0.0, 0.0, A[3, 1], A[3, 3]]], dtype = complex)

        r = np.dot(np.linalg.inv(r1), r2)
        self.r = r

        pBackward = np.array([self.layers[0].poynting[1], self.layers[0].poynting[3], \
                              self.layers[-1].poynting[1], self.layers[-1].poynting[3]])
 
        pForward = np.array([self.layers[0].poynting[0], self.layers[0].poynting[2], \
                              self.layers[-1].poynting[0], self.layers[-1].poynting[2]])
 
 
        R = np.zeros_like(r, dtype = complex)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            
            R[0, 0] = abs(r[0, 0]) ** 2.0 * abs(pBackward[0]) / abs(pForward[0])
            R[0, 1] = abs(r[0, 1]) ** 2.0 * abs(pBackward[0]) / abs(pForward[1])
            R[0, 2] = np.NAN #abs(r[0, 2]) ** 2.0 * abs(pBackward[0] / pBackward[2])
            R[0, 3] = np.NAN #abs(r[0, 3]) ** 2.0 * abs(pBackward[0] / pBackward[3])
            R[1, 0] = abs(r[1, 0]) ** 2.0 * abs(pBackward[1]) / abs(pForward[0])
            R[1, 1] = abs(r[1, 1]) ** 2.0 * abs(pBackward[1]) / abs(pForward[1])
            R[1, 2] = np.NAN #abs(r[1, 2]) ** 2.0 * abs(pBackward[1] / pBackward[2])
            R[1, 3] = np.NAN #abs(r[1, 3]) ** 2.0 * abs(pBackward[1] / pBackward[3])
            R[2, 0] = abs(r[2, 0]) ** 2.0 * abs(pForward[2]) / abs(pForward[0])
            R[2, 1] = abs(r[2, 1]) ** 2.0 * abs(pForward[2]) / abs(pForward[1])
            R[2, 2] = np.NAN #abs(r[2, 2]) ** 2.0 * abs(pForward[2] / pBackward[2])
            R[2, 3] = np.NAN #abs(r[2, 3]) ** 2.0 * abs(pForward[2] / pBackward[3])
            R[3, 0] = abs(r[3, 0]) ** 2.0 * abs(pForward[3]) / abs(pForward[0])
            R[3, 1] = abs(r[3, 1]) ** 2.0 * abs(pForward[3]) / abs(pForward[1])
            R[3, 2] = np.NAN #abs(r[3, 2]) ** 2.0 * abs(pForward[3] / pBackward[2])
            R[3, 3] = np.NAN #abs(r[3, 3]) ** 2.0 * abs(pForward[3] / pBackward[3])
            if len(w) > 0:
                print("warning", w)
                print(self.beta)
                print(pBackward)
                print(pForward)
     
        R = R.real
        self.R = R
        
        return r, R
    
    def CalcFields1D(self, xs, polarization = None):
        if polarization != None:
            self.polarization = polarization
        normCoef, coefsAll = self._CalcFieldCoefs()
        layerIndices, layerDs = self._LayerIndices(xs)
    
        resE = np.zeros((len(xs), 3), dtype = complex)
        resH = np.zeros((len(xs), 3), dtype = complex)
        
        for i in range(len(xs)):
            layerId = layerIndices[i]
            coefs = coefsAll[layerId, :]
            
            E, H = self.layers[layerId].GetFields(layerDs[i], coefs)
            resE[i, :] = E
            resH[i, :] = H
    
        resE /= normCoef
        resH /= normCoef
       
        return resE, resH
    
    def CalcEnhAtInterface(self, polarization = None, enhInterface = -1, enhDist = 0.0):
        if polarization != None:
            self.polarization = polarization
            
        if enhInterface < 0:
            layerId = len(self.layers) - 1
        else:
            layerId = enhInterface

        normCoef, coefsAll = self._CalcFieldCoefs()
        E, _ = self.layers[layerId].GetFields(enhDist, coefsAll[layerId, :])
        E /= normCoef
        
        enh = Norm(E)
        return enh, E
    
    def OptimizeForEnhancement(self, optParams, optInitial, polarization = None, enhInterface = -1, enhDist = 0.0):
        def FitFunc(x):
            
            for i in range(len(x)):
                self.SetConf(**{optParams[i]: x[i]})
            self.Solve()
            enh, _ = self.CalcEnhAtInterface(enhInterface = enhInterface, enhDist = enhDist)
            #print x, enh
            return -enh
        
        if polarization != None:
            self.polarization = polarization
        optValues, maxEnh, _, __, ___ = optimize.fmin(FitFunc, optInitial, disp = False, full_output = True)
        maxEnh *= -1
        #print "optValues", optValues
        for i in range(len(optParams)):
            self.SetConf(**{optParams[i]: optValues[i]})
        
        return optValues, maxEnh


    def _CalcFieldCoefs(self, polarization = None):
        if polarization != None:
            self.polarization = polarization
        a1In, a2In = self.polarization
        
        inputFields = np.array([a1In, a2In, 0.0, 0.0], dtype = complex)
        outputFields = np.dot(self.r, inputFields)
        
        Einc, _ = self.layers[0].GetFields(0.0, np.array([a1In, 0.0, a2In, 0.0]))
        
        normCoef = np.sqrt(abs(Einc[0]) ** 2.0 + abs(Einc[1]) ** 2.0 + abs(Einc[2]) ** 2.0, dtype = complex)
        n1 = np.sqrt(self.beta ** 2.0 + self.layers[0].alpha[0] ** 2.0)
        n2 = np.sqrt(self.beta ** 2.0 + self.layers[0].alpha[2] ** 2.0)
        nEff = (a1In * n1.real + a2In * n2.real) / (a1In + a2In) # Maybe not fully correct
        #print "n1", n1, "n2", n2, "nEff", nEff
        normCoef *= math.sqrt(nEff)
        
        coefsSubstrate = np.array([outputFields[2], 0.0, outputFields[3], 0.0])
        mat = self.layers[-1].F

        coefsAll = np.zeros((len(self.layers), 4), dtype = complex)
        
        

        for i in range(len(self.layers)-1, -1, -1):
            mat = np.dot(self.layers[i].M, mat)
            coefsAll[i, :] = np.dot(self.layers[i].invF, np.dot(mat, coefsSubstrate))

        coefsAll[-1, 1] = coefsAll[-1, 3] = 0.0  
        return normCoef, coefsAll
        
    
    def _LayerIndices(self, xs):
        resInd = np.zeros_like(xs, dtype = int)
        resD = np.zeros_like(xs, dtype = float)
        
        curLayer = 0
        curDist = 0.0
        prevDist = 0.0
        
        for i in range(len(xs)):
            while(xs[i] >= curDist):
                curLayer += 1
                prevDist = curDist
                if curLayer >= len(self.layers) - 1:
                    curDist = float("inf")
                    curLayer = len(self.layers) - 1
                curDist += self.layers[curLayer].d
                
            resInd[i] = curLayer
            resD[i] = xs[i] - prevDist
                
        return resInd, resD
        
     
if __name__ == "__main__":
    pass


import numpy as np
import pylab as plt
from GeneralTmm import Tmm, Material

if __name__ == "__main__":
    # Prepare materials
    prismN = Material.Static(1.5)
    substrateN = Material.Static(1.0)
    
    # Parameters
    wl = 532e-9
    betas = np.linspace(0.0, 1.49, 100)
    angles = np.arcsin(betas / prismN(wl).real)
    
    # Init TMM
    tmm = Tmm()
    tmm.wl = wl
    
    # Add layers
    tmm.AddIsotropicLayer(float("inf"), prismN)
    tmm.AddIsotropicLayer(float("inf"), substrateN)
    
    # Do calculations
    sr = tmm.Sweep("beta", betas)
    
    # Plot
    plt.figure(figsize = (8, 5))
    plt.suptitle("Total internal reflection")
    plt.subplot(121)
    plt.title("Reflection")
    plt.plot(np.degrees(angles), sr["R11"], label = r"p-pol")
    plt.plot(np.degrees(angles), sr["R22"], label = r"s-pol")
    plt.xlabel(r"$\theta$ ($\degree$)")
    plt.legend()
    
    plt.subplot(122)
    plt.title("Transmission")
    plt.plot(np.degrees(angles), sr["T31"], label = r"p-pol")
    plt.plot(np.degrees(angles), sr["T42"], label = r"s-pol")
    plt.xlabel(r"$\theta$ ($\degree$)")
    plt.legend()
    
    plt.show()


import re
import sys
import eigency
import numpy as np
from setuptools import setup
from setuptools.extension import Extension
from Cython.Build import cythonize


__version__ = re.search(r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
    open("GeneralTmm/__init__.py").read()).group(1)

extra_compile_args = []
if sys.platform in ("linux", "darwin"):
    extra_compile_args.append("-std=c++11")

extensions = cythonize([
    Extension(
        "GeneralTmm._GeneralTmmCppExt",
        sources=[
            "GeneralTmm/src/GeneralTmm.pyx",
            "GeneralTmm/src/Common.cpp",
            "GeneralTmm/src/Layer.cpp",
            "GeneralTmm/src/Material.cpp",
            "GeneralTmm/src/tmm.cpp",
        ],
        include_dirs=[
            np.get_include(),
            "GeneralTmm/src",
            "GeneralTmm/src/Simplex"
        ] + eigency.get_includes(),
        language="c++",
        extra_compile_args=extra_compile_args,
    ),
])

long_description = open("README.md").read()

setup(
    name="GeneralTmm",
    description="General 4x4 transfer-matric method (TMM)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=__version__,
    author="Ardi Loot",
    url="https://github.com/ardiloot/GeneralTmm",
    author_email="ardi.loot@outlook.com",
    packages=["GeneralTmm"],
    include_package_data=True,
    ext_modules=extensions,
    install_requires=["numpy", "scipy", "eigency"],
)



from libcpp cimport bool
from libcpp.map cimport map
from libcpp.vector cimport vector
from libcpp.string cimport string
from eigency.core cimport *

#===============================================================================
# Common.h
#===============================================================================
        
cdef extern from "Common.h" namespace "TmmModel":
    
    cdef enum ParamTypeCpp "TmmModel::ParamType":
        WL,
        BETA,
        ENH_OPT_REL,
        ENH_OPT_MAX_ITERS,
        ENH_INITIAL_STEP,
        LAYER_D,
        LAYER_N,
        LAYER_NX,
        LAYER_NY,
        LAYER_NZ,
        LAYER_PSI,
        LAYER_XI,
        LAYER_MAT_NX,
        LAYER_MAT_NY,
        LAYER_MAT_NZ,
        NOT_DEFINED
    
    #---------------------------------------------------------------------------
        
    cdef enum WaveDirectionCpp "TmmModel::WaveDirection":
        WD_FORWARD,
        WD_BACKWARD,
        WD_BOTH
    
    #---------------------------------------------------------------------------
    
    cdef cppclass ParamCpp "TmmModel::Param":
            ParamCpp() except +
            ParamCpp(ParamTypeCpp pType) except +
            ParamCpp(ParamTypeCpp pType_, int layerId_) except +
    
    #--------------------------------------------------------------------------- 
    
    cdef cppclass PositionSettingsCpp "TmmModel::PositionSettings":
        PositionSettingsCpp() except +
        PositionSettingsCpp(double polCoef1, double polCoef2, int interfaceId_, double distFromInterface_) except +
            
    #--------------------------------------------------------------------------- 

    cdef cppclass SweepResCpp "TmmModel::SweepRes":
        map[string, ArrayXcd] mapComplex
        map[string, ArrayXd] mapDouble
    
    #---------------------------------------------------------------------------
    
    cdef cppclass EMFieldsCpp "TmmModel::EMFields":
        ArrayXcd E
        ArrayXcd H
    
    #---------------------------------------------------------------------------
    
    cdef cppclass EMFieldsListCpp "TmmModel::EMFieldsList":
        MatrixXcd E
        MatrixXcd H
        
#===============================================================================
# Material.h
#===============================================================================
        
cdef extern from "Material.h" namespace "TmmModel":        
    
    cdef cppclass MaterialCpp "TmmModel::Material":
        MaterialCpp() except +
        MaterialCpp(Map[ArrayXd] & wlsExp, Map[ArrayXcd] & nsExp) except +
        double complex n(double wl) except +
            
#===============================================================================
# tmm.h
#===============================================================================
        
cdef extern from "tmm.h" namespace "TmmModel":

    cdef cppclass TmmCpp "TmmModel::Tmm":
        Tmm() except +
        
        void SetParam(ParamCpp ParamCpp, int value) except +
        void SetParam(ParamCpp ParamCpp, double value) except +
        void SetParam(ParamCpp ParamCpp, double complex value) except +
        
        int GetParamInt(ParamCpp ParamCpp) except +
        double GetParamDouble(ParamCpp ParamCpp) except + 
        double complex GetParamComplex(ParamCpp ParamCpp) except +
        
        void AddIsotropicLayer(double d, MaterialCpp *mat) except +
        void AddLayer(double d, MaterialCpp *matx, MaterialCpp *maty, MaterialCpp *matz, double psi, double xi) except +
        void ClearLayers() except +
        
        Matrix4d GetIntensityMatrix() except +
        Matrix4cd GetAmplitudeMatrix() except +
        SweepResCpp Sweep(ParamCpp sweepParamCpp, Map[ArrayXd] sweepValues, PositionSettingsCpp enhpos, int alphasLayer) except +
        
        EMFieldsListCpp CalcFields1D(Map[ArrayXd] xs, Map[Array2d] polarization, WaveDirectionCpp waveDirection) except +
        EMFieldsCpp CalcFieldsAtInterface(PositionSettingsCpp pos, WaveDirectionCpp waveDirection) except +
        double OptimizeEnhancement(vector[ParamCpp] optParams, Map[ArrayXd] optInitial, PositionSettingsCpp pos) except +
        




import unittest
import numpy as np
from GeneralTmm import Tmm, TmmPy, Material

# Refractive index data of silver from
# P. B. Johnson and R. W. Christy. Optical Constants of the Noble Metals, Phys. Rev. B 6, 4370-4379 (1972)

wlsAg = np.array([4.0e-07,   4.2e-07,   4.4e-07,   4.6e-07, 4.8e-07,   5.0e-07,
    5.2e-07,   5.4e-07, 5.6e-07,   5.8e-07,   6.0e-07,   6.2e-07, 6.4e-07,
    6.6e-07,   6.8e-07,   7.0e-07, 7.2e-07,   7.4e-07,   7.6e-07,   7.8e-07,
    8.0e-07,   8.2e-07,   8.4e-07,   8.6e-07, 8.8e-07,   9.0e-07,   9.2e-07,
    9.4e-07, 9.6e-07,   9.8e-07])
    
nsAg = np.array([0.050 + 2.104j, 0.046 + 2.348j, 0.040 + 2.553j, 0.044 + 2.751j,
    0.050 + 2.948j, 0.050 + 3.131j, 0.050 + 3.316j, 0.057 + 3.505j, 0.057 + 3.679j,
    0.051 + 3.841j, 0.055 + 4.010j, 0.059 + 4.177j, 0.055 + 4.332j, 0.050 + 4.487j,
    0.045 + 4.645j, 0.041 + 4.803j, 0.037 + 4.960j, 0.033 + 5.116j, 0.031 + 5.272j,
    0.034 + 5.421j, 0.037 + 5.570j, 0.040 + 5.719j, 0.040 + 5.883j, 0.040 + 6.048j,
    0.040 + 6.213j, 0.040 + 6.371j, 0.040 + 6.519j, 0.040 + 6.667j, 0.040 + 6.815j,
    0.040 + 6.962j])

class TmmTest(unittest.TestCase):
    
    @classmethod
    def _PrepareTmm(self, wl, layers):
        tmm = Tmm();
        tmm.SetParams(wl = wl)
        for layer in layers:
            if layer[0] == "iso":
                _, d, n = layer
                mat = Material.Static(n)
                tmm.AddIsotropicLayer(d, mat)
            else:
                _, d, n1, n2, n3, psi, xi = layer
                matx = Material.Static(n1)
                maty = Material.Static(n2)
                matz = Material.Static(n3)
                tmm.AddLayer(d, matx, maty, matz, psi, xi)

        # Old Tmm
        oldTmm = TmmPy()
        oldTmm.SetConf(wl = wl)
        for layer in layers:
            if layer[0] == "iso":
                _, d, n = layer
                oldTmm.AddIsotropicLayer(d, n)
            else:
                _, d, n1, n2, n3, psi, xi = layer
                oldTmm.AddLayer(d, n1, n2, n3, psi, xi)
                
        return tmm, oldTmm
        
    @classmethod
    def _testSweep(self, wl, layers, betas, pol, enhInterface, enhDist):
        tmm, oldTmm = self._PrepareTmm(wl, layers)
        res = tmm.Sweep("beta", betas, (pol, enhInterface, enhDist))
        resOld = oldTmm.SolveFor("beta", betas, polarization = pol, \
                                 enhInterface = enhInterface, enhDist = enhDist)
        for k in ["R11", "R22", "R12", "R21", "T31", "T32", "T41", "T42"]:
            np.testing.assert_almost_equal(res[k], resOld[k])
            
        np.testing.assert_almost_equal(res["enh"], resOld["enh"])
    
    @classmethod
    def _testFields(self, wl, beta, layers, xs, pol):
        tmm, oldTmm = self._PrepareTmm(wl, layers)
        
        tmm.beta = beta
        E, H = tmm.CalcFields1D(xs, np.array(pol))
        oldTmm.Solve(wl, beta)
        EOld, HOld = oldTmm.CalcFields1D(xs, pol)

        for i in range(3):
            print("test E", "XYZ"[i])
            np.testing.assert_allclose(abs(E[:, i]), abs(EOld[:, i]), rtol = 1e-7, atol = 1e-15)
            
        for i in range(3):
            print("test H", "XYZ"[i])     
            np.testing.assert_allclose(abs(H[:, i]), abs(HOld[:, i]), rtol = 1e-7, atol = 1e-15)

            
    def testTIRSweep(self):
        for wl in [400e-9, 800e-9]:
            for n1 in [1.2, 1.5]:
                for n2 in [1.0, 2.0]:
                    for pol in [(1.0, 0.0), (0.0, 1.0)]:
                        print("TIR test", wl, n1, n2, pol)
                        layers = [("iso", float("inf"), n1), \
                                  ("iso", float("inf"), n2)]
                        betas = np.linspace(0.0, n1 - 1e-3, 30)
                        self._testSweep(wl, layers, betas, pol, -1, 0.0)

    def testSppSweep(self):
        silver = Material(wlsAg, nsAg)
        for wl in [400e-9, 500e-9, 800e-9]:
            for pol in [(1.0, 0.0), (0.0, 1.0)]:
                print("SPP test", wl, pol)
                layers = [("iso", float("inf"), 1.5), \
                          ("iso", float("inf"), silver(wl)),
                          ("iso", float("inf"), 1.0)]
                betas = np.linspace(0.0, 1.499, 30)
                self._testSweep(wl, layers, betas, pol, -1, 0.0)

    def testAnisoSweep(self):
        wl = 500e-9
        for pol in [(1.0, 0.0), (0.0, 1.0)]:
            for psi in np.linspace(0.0, 2 * np.pi, 5):
                for xi in np.linspace(0.0, np.pi, 5):
                    print("Aniso test", pol, psi, xi)
                    layers = [("iso", float("inf"), 1.5), \
                              ("aniso", float("inf"), 1.76, 1.8, 1.9, psi, xi)]
                    betas = np.linspace(0.0, 1.5 - 1e-3, 30)
                    self._testSweep(wl, layers, betas, pol, -1, 0.0)
                    
    def testAnisoSweep2(self):
        wl = 500e-9
        for pol in [(1.0, 0.0), (0.0, 1.0)]:
            for psi in np.linspace(0.0, 2 * np.pi, 5):
                for xi in np.linspace(0.0, np.pi, 5):
                    print("Aniso test", pol, psi, xi)
                    layers = [("iso", float("inf"), 1.5), \
                              ("aniso", 200e-9, 1.76, 1.8, 1.9, psi, xi),
                              ("iso", float("inf"), 1.5)]
                    betas = np.linspace(0.0, 1.5 - 1e-3, 30)
                    self._testSweep(wl, layers, betas, pol, -1, 0.0)


    def testSppFields(self):
        silver = Material(wlsAg, nsAg)
        beta = 0.5
        xs = np.linspace(-1e-6, 1e-6, 1000)
        for wl in [400e-9, 500e-9, 800e-9]:
            for pol in [(1.0, 0.0), (0.0, 1.0)]:
                print("SPP test", wl, pol)
                layers = [("iso", float("inf"), 1.5), \
                          ("iso", float("inf"), silver(wl)),
                          ("iso", float("inf"), 1.0)]
                self._testFields(wl, beta, layers, xs, pol)
                
    def testAnisoFields(self):
        beta = 0.5
        xs = np.linspace(-1e-6, 1e-6, 100)
        for wl in [400e-9, 500e-9, 800e-9]:
            for psi in np.linspace(0.0, 2 * np.pi, 5):
                for xi in np.linspace(0.0, np.pi, 5):
                    for pol in [(1.0, 0.0), (0.0, 1.0)]:
                        print(wl, psi, xi, pol)
                        layers = [("iso", float("inf"), 1.5), \
                                  ("aniso", float("inf"), 1.76, 1.8, 1.9, psi, xi)]
                        self._testFields(wl, beta, layers, xs, pol)

if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TmmTest)
    unittest.TextTestRunner().run(suite)